"""Models for node label prediction."""
from .node_label_prediction_feed_forward_neural_network import NodeLabelPredictionfeedForwardNeuralNetwork
from .graph_convolutional_neural_networks import GraphConvolutionalNeuralNetwork

__all__ = [
    "NodeLabelPredictionfeedForwardNeuralNetwork",
    "GraphConvolutionalNeuralNetwork"
]



"""Keras Sequence for running Neural Network on graph edge prediction."""
from typing import Union, Tuple, List
import numpy as np

from ensmallen import Graph  # pylint: disable=no-name-in-module
from .edge_prediction_sequence import EdgePredictionSequence


class EdgeLabelPredictionSequence(EdgePredictionSequence):
    """Keras Sequence for running a Neural Network on graph edge-label prediction."""

    def __init__(
        self,
        graph: Graph,
        use_node_types: bool = False,
        use_edge_metrics: bool = False,
        batch_size: int = 2**10,
        support_mirrored_strategy: bool = False,
        graph_to_avoid: Graph = None,
        batches_per_epoch: Union[int, str] = "auto",
        elapsed_epochs: int = 0,
        random_state: int = 42
    ):
        """Create new EdgeLabekPredictionSequence object.

        Parameters
        --------------------------------
        graph: Graph,
            The graph from which to sample the edges.
        use_node_types: bool = False,
            Whether to return the node types.
        use_edge_metrics: bool = False,
            Whether to return the edge metrics.
        batch_size: int = 2**10,
            The batch size to use.
        support_mirrored_strategy: bool = False,
            Wethever to patch support for mirror strategy.
            At the time of writing, TensorFlow's MirrorStrategy does not support
            input values different from floats, therefore to support it we need
            to convert the unsigned int 32 values that represent the indices of
            the embedding layers we receive from Ensmallen to floats.
            This will generally slow down performance, but in the context of
            exploiting multiple GPUs it may be unnoticeable.
        graph_to_avoid: Graph = None,
            Graph to avoid when generating the edges.
            This can be the validation component of the graph, for example.
            More information to how to generate the holdouts is available
            in the Graph package.
        batches_per_epoch: Union[int, str] = "auto",
            Number of batches per epoch.
            If auto, it is used: `10 * edges number /  batch size`
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        random_state: int = 42,
            The random_state to use to make extraction reproducible.
        """
        super().__init__(
            graph,
            use_node_types=use_node_types,
            use_edge_types=True,
            return_only_edges_with_known_edge_types=True,
            use_edge_metrics=use_edge_metrics,
            batch_size=batch_size,
            negative_samples_rate=0,
            avoid_false_negatives=False,
            support_mirrored_strategy=support_mirrored_strategy,
            graph_to_avoid=graph_to_avoid,
            batches_per_epoch=batches_per_epoch,
            elapsed_epochs=elapsed_epochs,
            random_state=random_state
        )

    def __getitem__(self, idx: int) -> Tuple[List[np.ndarray], np.ndarray]:
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        Return Tuple containing X and Y numpy arrays corresponding to given batch index.
        """
        edge_data, _ = super().__getitem__(idx)
        return edge_data[:-1], edge_data[-1]



"""Module implementing edge and edge-label prediction models."""
from .perceptron import Perceptron
from .multi_layer_perceptron import MultiLayerPerceptron
from .feed_forward_neural_network import FeedForwardNeuralNetwork
from .edge_prediction_graph_neural_network import EdgePredictionGraphNeuralNetwork

__all__ = [
    "Perceptron",
    "MultiLayerPerceptron",
    "FeedForwardNeuralNetwork",
    "EdgePredictionGraphNeuralNetwork"
]



"""Keras Sequence for running GNN on graph edge prediction."""
from typing import Tuple, Union, List, Optional

import numpy as np
import pandas as pd
from ensmallen import Graph  # pylint: disable=no-name-in-module
from keras_mixed_sequence import VectorSequence
from tqdm.auto import tqdm


class GNNBipartiteEdgePredictionSequence(VectorSequence):
    """Keras Sequence for running GNN on graph edge prediction."""

    def __init__(
        self,
        graph: Graph,
        sources: np.ndarray,
        destinations: np.ndarray,
        node_features: Optional[Union[pd.DataFrame, List[pd.DataFrame], np.ndarray, List[np.ndarray]]] = None,
        use_node_types: bool = False,
        return_node_ids: bool = True,
        return_labels: bool = True,
        support_mirrored_strategy: bool = False,
        random_state: int = 42
    ):
        """Create new GNNEdgePredictionSequence object.

        Parameters
        --------------------------------
        TODO: Update
        """
        self._sources = sources
        self._destinations = destinations
        self._graph = graph

        self._source_node_type_ids = None
        self._destination_node_type_ids = None

        if use_node_types:
            self._source_node_type_ids = np.zeros((
                sources.size,
                graph.get_maximum_multilabel_count()
            ), dtype=np.uint32)
            for i, source_node_id in enumerate(tqdm(
                sources,
                desc="Computing source node types",
                leave=False,
                dynamic_ncols=True
            )):
                node_type_ids = graph.get_node_type_ids_from_node_id(
                    source_node_id
                )
                for j, node_type_id in enumerate(node_type_ids):
                    self._source_node_type_ids[i, j] = node_type_id + 1
            self._destination_node_type_ids = np.zeros((
                destinations.size,
                graph.get_maximum_multilabel_count()
            ), dtype=np.uint32)
            for i, destination_node_id in enumerate(tqdm(
                destinations,
                desc="Computing destination node types",
                leave=False,
                dynamic_ncols=True
            )):
                node_type_ids = graph.get_node_type_ids_from_node_id(
                    destination_node_id
                )
                for j, node_type_id in enumerate(node_type_ids):
                    # Here we need to offset the node type IDs by one
                    # because we use the node type zero as the padding
                    self._destination_node_type_ids[i, j] = node_type_id + 1

        self._node_features = [
            node_feature.values
            if isinstance(node_feature, pd.DataFrame)
            else node_feature
            for node_feature in node_features
        ]
        self._destination_node_features = [
            node_feature[self._destinations]
            for node_feature in self._node_features
        ]
        self._use_node_types = use_node_types
        self._return_node_ids = return_node_ids
        self._return_labels = return_labels

        if self._return_node_ids:
            self._destination_node_features.append(self._destinations)
        if self._use_node_types:
            self._destination_node_features.append(
                self._destination_node_type_ids)
        self._support_mirrored_strategy = support_mirrored_strategy
        super().__init__(
            sources,
            batch_size=1,
            random_state=random_state
        )

    def __getitem__(self, idx: int) -> List[np.ndarray]:
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        Return Tuple containing X and Y numpy arrays corresponding to given batch index.
        """
        sources = np.full_like(
            self._destinations,
            super().__getitem__(idx)[0]
        )
        source_node_types = None
        if self._source_node_type_ids is not None:
            source_node_types = np.tile(
                self._source_node_type_ids[idx],
                (self._destinations.size, 1)
            )

        source_ids = None
        if self._return_node_ids:
            source_ids = sources

        X = [
            node_feature[sources]
            for node_feature in self._node_features
        ] + [
            value
            for value in (
                source_ids, source_node_types
            )
            if value is not None
        ] + self._destination_node_features

        if self._return_labels:
            labels = np.fromiter(
                (
                    self._graph.has_edge_from_node_ids(
                        source_id,
                        destination_id
                    )
                    for source_id, destination_id in zip(
                        sources,
                        self._destinations
                    )
                ),
                dtype=np.bool
            )
            return X, labels
        return X
